//
//  ConvexHull2d.h
//  ConvexHull2d
//
//  Created by kumada on 2013/06/02.
//  Copyright (c) 2013年 kumada. All rights reserved.
//

#ifndef ConvexHull2d_ConvexHull2d_h
#define ConvexHull2d_ConvexHull2d_h

#include "VertexExtractor.h"
#include "SymbolicPerturbation2d.h"
#include "Point.h"
#include "Debug.h"
#include "Concepts.h"
#include <boost/range/algorithm/sort.hpp>
#include <boost/range/algorithm/for_each.hpp>
#include <boost/iterator/counting_iterator.hpp>

/** generate the convex hull of points in the 2-dimensional space
        @tparam T int or variable-length integer
*/
template<typename T>
class ConvexHull2d
{
        friend class ConvexHull2dTester;
        BOOST_CLASS_REQUIRE(T, geo, IntegerConcept);
        
public:
        typedef geo::Point<T, geo::TwoDimension> Point;
        typedef std::unique_ptr<Point>           PointPtr;

        // test ok
        explicit ConvexHull2d(const std::vector<PointPtr>& points)
                : original_points_(points)
                , sorted_to_original_indices_(generate_names(points))
                , simbolic_perturbation_{points, sorted_to_original_indices_}
                , ccvertices_(points.size(), -1)
                , cvertices_(points.size(), -1)
                , previous_index_{} {}
        
        ConvexHull2d() = default;
        ~ConvexHull2d() = default;
        
        void execute()
        {
                generate_first_triangle();
                previous_index_ = 2;
                
                DEBUG_SEE_VERTICES(int, ccvertices_, original_points_, sorted_to_original_indices_);
                for ( int i = 3, n = static_cast<int>(sorted_to_original_indices_.size()); i < n; ++i ) {
                        add_vertex(i);
                        DEBUG_SEE_VERTICES(int, ccvertices_, original_points_, sorted_to_original_indices_);
                }
        }

        const VertexExtractor<T> extract_ccvertices() const
        {
                return VertexExtractor<T> {original_points_, sorted_to_original_indices_, ccvertices_};
        }
        
        const VertexExtractor<T> extract_cvertices() const
        {
                return VertexExtractor<T> {original_points_, sorted_to_original_indices_, cvertices_};
        }

private:
        void generate_first_triangle()
        {
                // pass the sorted numbers
                bool turns_left = simbolic_perturbation_.turns_left(0, 1, 2);
                
                if ( turns_left ) { // counter clockwise
                        ccvertices_[0] = 1;
                        ccvertices_[1] = 2;
                        ccvertices_[2] = 0;

                        cvertices_[0] = 2;
                        cvertices_[1] = 0;
                        cvertices_[2] = 1;
                        
                } else { // clockwise
                        ccvertices_[0] = 2;
                        ccvertices_[1] = 0;
                        ccvertices_[2] = 1;

                        cvertices_[0] = 1;
                        cvertices_[1] = 2;
                        cvertices_[2] = 0;
                }
        }

        ConvexHull2d(const ConvexHull2d&) = delete;
        ConvexHull2d& operator=(const ConvexHull2d&) = delete;
        ConvexHull2d(ConvexHull2d&&) = delete;
        ConvexHull2d& operator=(ConvexHull2d&&) = delete;
        
        const std::vector<PointPtr>& original_points_;
        std::vector<int>             sorted_to_original_indices_;
        SymbolicPerturbation2d<T>    simbolic_perturbation_;
        std::vector<int>             ccvertices_;
        std::vector<int>             cvertices_;
        int                          previous_index_;
        
        /** pass the sorted number
                @param[in] current_index sorted number
        */
        void add_vertex(const int current_index)
        {
                std::function<bool(bool)> f = [](bool returns_left){ return returns_left; };
                int p0 = trace(current_index, previous_index_, ccvertices_, f);
                int p1 = trace(current_index, previous_index_, cvertices_, std::not1(f));
                
                if ( p0 != p1 ) {
                        remove_vertices(p1, p0, ccvertices_);
                        remove_vertices(p0, p1, cvertices_);
                
                        add_vertices(p1, p0, current_index, ccvertices_);
                        add_vertices(p0, p1, current_index, cvertices_);
                        previous_index_ = current_index;
                }
        }
        
        // test ok
        void add_vertices(int first, int last, int index, std::vector<int>& vertices) const
        {
                vertices[first] = index;
                vertices[index] = last;
        }
        
        // test ok
        void remove_vertices(int first, int last, std::vector<int>& vertices)
        {
                int index = vertices[first];
                while ( true ) {
                        if ( index == last ) {
                                break;
                        }
                        int tmp = vertices[index];
                        vertices[index] = -1;
                        index = tmp;
                }
        }

        // test ok
        template<typename F>
        int trace(int current_index, int index, const std::vector<int>& vertices, F fun) const
        {
                while ( true ) {
                        bool turns_left = simbolic_perturbation_.turns_left(current_index, index, vertices[index]);
                        if ( fun(turns_left) ) {
                                break;
                        }
                        index = vertices[index];
                }
                return index;
        }
        
        std::vector<int> generate_names(const std::vector<PointPtr>& points) const
        {
                if ( points.empty() ) {
                        throw std::runtime_error("input is empty");
                }

                std::vector<int> names(
                        boost::counting_iterator<int>(0),
                        boost::counting_iterator<int>(static_cast<int>(points.size()))
                );
                
                boost::sort(names, [&](int lhs, int rhs){
                        const T lhsx = points[lhs]->at(0);
                        const T rhsx = points[rhs]->at(0);
                        return (lhsx == rhsx) ? lhs > rhs : lhsx < rhsx;
                });
                return names; // move or RVO
        }
};
#endif
