//
//  ConvexHull2d.cpp
//  ConvexHull2d
//
//  Created by kumada on 2013/06/02.
//  Copyright (c) 2013年 kumada. All rights reserved.
//
#if(UNIT_TEST)
#define BOOST_TEST_DYN_LINK

#include "ConvexHull2d.h"
#include <memory>
#include <random>
#include <boost/test/unit_test.hpp>
#include <boost/multiprecision/cpp_int.hpp>
#include <boost/range/algorithm/sort.hpp>
#include <boost/range/algorithm/transform.hpp>

namespace mp = boost::multiprecision;

class ConvexHull2dTester
{
public:
        template<typename T>
        static std::vector<typename ConvexHull2d<T>::PointPtr> create_random_points(int n, int first, int last)
        {
                std::default_random_engine e {};
                std::uniform_int_distribution<int> d {first, last};
                typedef typename ConvexHull2d<T>::Point Point;
                typedef typename ConvexHull2d<T>::PointPtr PointPtr;
                
                std::vector<PointPtr> points(n);
                for ( auto& point : points ) {
                        point = PointPtr(new Point{{d(e), d(e)}});
                }
                return points;
        }

        template<typename T>
        static std::vector<typename ConvexHull2d<T>::PointPtr> create_random_degenerate_points()
        {
                const T ps[][2] = {
                        {1, 166},
                        {1, 861},
                        {23, -620},
                        {13, -760},
                        {1, -710},
                        {1, -408},
                        {1, -912},
                        {1, -806},
                        {1, -362},
                        {1, 314},
                };

                typedef typename ConvexHull2d<T>::Point Point;
                typedef typename ConvexHull2d<T>::PointPtr PointPtr;
                
                std::vector<PointPtr> points(10);
                for ( int i = 0; i < sizeof(ps) / sizeof(ps[0]); ++i ) {
                        points[i] = PointPtr(new Point{{ps[i][0], ps[i][1]}});
                }
                return points;
        }

        template<typename T>
        static void test_constructor()
        {
                auto points = create_random_degenerate_points<T>();

                ConvexHull2d<T> ch2d{points};
                const auto& indices = ch2d.sorted_to_original_indices_;
                int i = 0;
                const T answers[] = {9, 8, 7, 6, 5, 4, 1, 0, 3, 2};
                for ( auto& index : indices ) {
                        BOOST_CHECK_EQUAL(answers[i++], index);
                }

                for ( std::size_t i = 0; i < indices.size() - 1; ++i ) {
                        const auto& p0 = points[indices[i]];
                        const auto& p1 = points[indices[i + 1]];
                        BOOST_CHECK(p0->at(0) <= p1->at(0));
                }
        }
        
        template<typename T>
        static void test_constructor_with_exception()
        {
                try {
                        std::vector<typename ConvexHull2d<T>::PointPtr> points;
                        ConvexHull2d<T> ch2d{points};
                        BOOST_CHECK_EQUAL(1, 0);
                } catch (const std::runtime_error& e) {
                        BOOST_CHECK_EQUAL("input is empty", e.what());
                }
        }
        
        template<typename T>
        static void test_trace()
        {
                std::vector<typename ConvexHull2d<T>::PointPtr> points(4);
                points[0] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{3, 86}}));
                points[1] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{14, 97}}));
                points[2] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{14, 78}}));
                points[3] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{40, 2}}));

                ConvexHull2d<T> ch2d{points};
                const T answers[] = {0, 2, 1, 3, 4};
                auto& names = ch2d.sorted_to_original_indices_;
                for ( std::size_t i = 0; i < names.size(); ++i ) {
                        BOOST_CHECK_EQUAL(answers[i], names[i]);
                }
                auto& ccvertices = ch2d.ccvertices_;
                ccvertices[0] = 2;
                ccvertices[1] = 0;
                ccvertices[2] = 1;
                
                auto& cvertices = ch2d.cvertices_;
                cvertices[0] = 1;
                cvertices[1] = 2;
                cvertices[2] = 0;
                
                std::function<bool(bool)> f = [](bool flag){ return flag; };
                BOOST_CHECK_EQUAL(2, ch2d.trace(3, 2, ccvertices, f));
                BOOST_CHECK_EQUAL(0, ch2d.trace(3, 2, cvertices, std::not1(f)));
        }

        template<typename T>
        static void test_trace_with_vertical_degeneracy()
        {
                std::vector<typename ConvexHull2d<T>::PointPtr> points(4);
                points[0] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{3, 86}}));
                points[1] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{3, 97}}));
                points[2] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{3, 78}}));
                points[3] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{3, 2}}));

                ConvexHull2d<T> ch2d{points};
                const T answers[] = {3, 2, 1, 0};
                auto& indices = ch2d.sorted_to_original_indices_;
                for ( std::size_t i = 0; i < indices.size(); ++i ) {
                        BOOST_CHECK_EQUAL(answers[i], indices[i]);
                }
                auto& ccvertices = ch2d.ccvertices_;
                ccvertices[0] = 1;
                ccvertices[1] = 2;
                ccvertices[2] = 0;
                
                auto& cvertices = ch2d.cvertices_;
                cvertices[0] = 2;
                cvertices[1] = 0;
                cvertices[2] = 1;
                
                std::function<bool(bool)> f = [](bool flag){ return flag; };
                BOOST_CHECK_EQUAL(2, ch2d.trace(3, 2, ccvertices, f));
                BOOST_CHECK_EQUAL(0, ch2d.trace(3, 2, cvertices, std::not1(f)));
        }
        
        template<typename T>
        static void test_trace_with_horizontal_degeneracy()
        {
                std::vector<typename ConvexHull2d<T>::PointPtr> points(4);
                points[0] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{3, 1}}));
                points[1] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{14, 1}}));
                points[2] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{15, 1}}));
                points[3] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{40, 1}}));

                ConvexHull2d<T> ch2d{points};
                auto& names = ch2d.sorted_to_original_indices_;
                for ( std::size_t i = 0; i < names.size(); ++i ) {
                        BOOST_CHECK_EQUAL(i, names[i]);
                }
                auto& ccvertices = ch2d.ccvertices_;
                ccvertices[0] = 1;
                ccvertices[1] = 2;
                ccvertices[2] = 0;
                
                auto& cvertices = ch2d.cvertices_;
                cvertices[0] = 2;
                cvertices[1] = 0;
                cvertices[2] = 1;
                
                std::function<bool(bool)> f = [](bool flag){ return flag; };
                BOOST_CHECK_EQUAL(0, ch2d.trace(3, 2, ccvertices, f));
                BOOST_CHECK_EQUAL(2, ch2d.trace(3, 2, cvertices, std::not1(f)));
        }
        
        template<typename T>
        static void test_remove_vertices()
        {
                std::vector<typename ConvexHull2d<T>::PointPtr> points(4);
                
                // dummy
                points[0] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{3, 1}}));
                points[1] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{14, 1}}));
                points[2] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{15, 1}}));
                points[3] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{40, 1}}));

                ConvexHull2d<T> ch2d{points};
                auto& names = ch2d.sorted_to_original_indices_;
                for ( std::size_t i = 0; i < names.size(); ++i ) {
                        BOOST_CHECK_EQUAL(i, names[i]);
                }

                auto& ccvertices = ch2d.ccvertices_;
                ccvertices[0] = 3;
                ccvertices[1] = 0;
                ccvertices[2] = 1;
                ccvertices[3] = 2;
                
                auto& cvertices = ch2d.cvertices_;
                cvertices[0] = 1;
                cvertices[1] = 2;
                cvertices[2] = 3;
                cvertices[3] = 0;
                
                ch2d.remove_vertices(3, 0, ccvertices);
                BOOST_CHECK_EQUAL(3, ccvertices[0]);
                BOOST_CHECK_EQUAL(-1, ccvertices[1]);
                BOOST_CHECK_EQUAL(-1, ccvertices[2]);
                BOOST_CHECK_EQUAL(2, ccvertices[3]);
                
                ch2d.remove_vertices(1, 3, cvertices);
                BOOST_CHECK_EQUAL(1, cvertices[0]);
                BOOST_CHECK_EQUAL(2, cvertices[1]);
                BOOST_CHECK_EQUAL(-1, cvertices[2]);
                BOOST_CHECK_EQUAL(0, cvertices[3]);
        }
        
        template<typename T>
        static void test_add_vertices()
        {
                std::vector<typename ConvexHull2d<T>::PointPtr> points(4);
                
                // dummy
                points[0] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{3, 1}}));
                points[1] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{14, 1}}));
                points[2] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{15, 1}}));
                points[3] = typename ConvexHull2d<T>::PointPtr(new typename ConvexHull2d<T>::Point({{40, 1}}));

                ConvexHull2d<T> ch2d{points};
                auto& names = ch2d.sorted_to_original_indices_;
                for ( std::size_t i = 0; i < names.size(); ++i ) {
                        BOOST_CHECK_EQUAL(i, names[i]);
                }
                std::vector<int> vertices = {0, 1, -1};
                ch2d.add_vertices(0, 1, 2, vertices);
                BOOST_CHECK_EQUAL(vertices[0], 2);
                BOOST_CHECK_EQUAL(vertices[2], 1);
                BOOST_CHECK_EQUAL(vertices[1], 1);
        }

};

BOOST_AUTO_TEST_CASE(TEST_ConvexHull2d)
{
        std::cout << "ConvexHull2d\n";
        
        ConvexHull2dTester::test_constructor<int>();
        ConvexHull2dTester::test_constructor<mp::cpp_int>();

        ConvexHull2dTester::test_constructor_with_exception<int>();
        ConvexHull2dTester::test_constructor_with_exception<mp::cpp_int>();

        ConvexHull2dTester::test_trace<int>();
        ConvexHull2dTester::test_trace<mp::cpp_int>();

        ConvexHull2dTester::test_trace_with_vertical_degeneracy<int>();
        ConvexHull2dTester::test_trace_with_vertical_degeneracy<mp::cpp_int>();

        ConvexHull2dTester::test_trace_with_horizontal_degeneracy<int>();
        ConvexHull2dTester::test_trace_with_horizontal_degeneracy<mp::cpp_int>();

        ConvexHull2dTester::test_remove_vertices<int>();
        ConvexHull2dTester::test_remove_vertices<mp::cpp_int>();
        
        ConvexHull2dTester::test_add_vertices<int>();
        ConvexHull2dTester::test_add_vertices<mp::cpp_int>();
}
#endif // UNIT_TEST
