//
//  SymbolicPerturbation2d.h
//  SymbolicPerturbation2d
//
//  Created by kumada on 2013/05/31.
//  Copyright (c) 2013年 kumada. All rights reserved.
//

#ifndef SymbolicPerturbation2d_SymbolicPerturbation2d_h
#define SymbolicPerturbation2d_SymbolicPerturbation2d_h

#include <vector>
#include "Point.h"
#include "Concepts.h"

/** execute the simbolic perturbation in the 2-dimensional space

        @tparam T int or variable-length integer
        @see Keisan Kika programming, Atsuyoshi Sugiura, p75
*/
template<typename T>
class SymbolicPerturbation2d
{
        friend class SymbolicPerturbation2dTester;
        BOOST_CLASS_REQUIRE(T, geo, IntegerConcept);
        
public:
        typedef geo::Point<T, geo::TwoDimension> Point;
        typedef std::unique_ptr<Point>           PointPtr;
        
        SymbolicPerturbation2d() = default;
        SymbolicPerturbation2d(
                const std::vector<PointPtr>& original_points,
                const std::vector<int>&      names
        )
                : original_points_(original_points)
                , sorted_to_original_indices_(names) {}
        
        ~SymbolicPerturbation2d() = default;
        
        // test ok
        /**
                @param[in] i sorted indices, i < j < k
                @param[in] j
                @param[in] k
        */
        bool turns_left(int i, int j, int k) const
        {
                // convert to the original indices
                const std::array<int, 3> ilist = {
                        sorted_to_original_indices_[i],
                        sorted_to_original_indices_[j],
                        sorted_to_original_indices_[k]
                };
                
                std::array<int, 3> jlist = {};
                auto is_even_permutation = sort(ilist, jlist);
                
                const auto& p0 = original_points_[jlist[0]];
                const auto& p1 = original_points_[jlist[1]];
                const auto& p2 = original_points_[jlist[2]];
                
                auto sign = compute_sign_of_determinant(p0, p1, p2);
                return judge(sign, is_even_permutation);
        }
        
private:
        const std::vector<PointPtr>& original_points_;
        const std::vector<int>&      sorted_to_original_indices_;

        SymbolicPerturbation2d(const SymbolicPerturbation2d&) = delete;
        SymbolicPerturbation2d& operator=(const SymbolicPerturbation2d&) = delete;
        SymbolicPerturbation2d(SymbolicPerturbation2d&&) = delete;
        SymbolicPerturbation2d& operator=(SymbolicPerturbation2d&&) = delete;

        int compute_sign_of_determinant(const PointPtr& p0, const PointPtr& p1, const PointPtr& p2) const
        {
                // first trial
                const std::array<T, 3> ia1 = {p0->at(0), p1->at(0), p2->at(0)};
                const std::array<T, 3> ia2 = {p0->at(1), p1->at(1), p2->at(1)};
                auto sign = compute_sign_of_determinant(ia1, ia2);
                if ( sign != 0 ) {
                        return sign;
                }

                // degenerate case(sign == 0)
                // y: i,j
                sign = -compute_sign_of_determinant(p0->at(1), p1->at(1));
                if ( sign != 0 ) {
                        return sign;
                }

                // degenerate case(sign == 0)
                // y: i,k
                sign = compute_sign_of_determinant(p0->at(1), p2->at(1));
                if ( sign != 0 ) {
                        return sign;
                }
                
                // degenerate case(sign == 0)
                // y: j,k
                sign = -compute_sign_of_determinant(p1->at(1), p2->at(1));
                if ( sign != 0 ) {
                        return sign;
                }
                
                // degenerate case(sign == 0)
                // x: i,j
                sign = compute_sign_of_determinant(p0->at(0), p1->at(0));
                if ( sign != 0 ) {
                        return sign;
                }

                // degenerate case(sign == 0)
                // x: i,k
                sign = -compute_sign_of_determinant(p0->at(0), p2->at(0));
                if ( sign != 0 ) {
                        return sign;
                }
                
                // degenerate case(sign == 0)
                // x: j,k
                sign = compute_sign_of_determinant(p1->at(0), p2->at(0));
                if ( sign != 0 ) {
                        return sign;
                }
                
                // last trial
                return 1;
        }
        
        // test ok
        /** sort "ilist" and store the result to "jlist"
         
          @param[in]  ilist input
          @param[out] jlist output
          @retval     true  if "jlist" is an even permutation of "ilist"
          @retval     false otherwise
        */
        bool sort(const std::array<int, 3>& ilist, std::array<int, 3>& jlist) const
        {
                jlist = ilist;
                auto& a = jlist[0];
                auto& b = jlist[1];
                auto& c = jlist[2];
 
                int s = 1;
                // a vs b,c
                if ( a < b ) {
                        std::swap(a, b);
                        s *= -1;
                }
                
                if ( a < c ) {
                        std::swap(a, c);
                        s *= -1;
                }
                
                // b vs c
                if ( b < c ) {
                        std::swap(b, c);
                        s *= -1;
                }
                
                return (s == 1) ? true : false;
        }
        
        // test ok
        /** compute the sign of the determinant of the form
                1 ia1[0] ia2[0]
                1 ia1[1] ia2[1]
                1 ia1[2] ia2[2]
                
         @param[in] ia1
         @param[in] ia2
         @retval      1 positive
         @retval     -1 negative
         @retval      0 zero(degenerate)
        */
        int compute_sign_of_determinant(
                const std::array<T, 3>& ia1,
                const std::array<T, 3>& ia2
        ) const {
                /*
                        1             0             0
                        0 ia1[1]-ia1[0] ia2[1]-ia2[0]
                        0 ia1[2]-ia1[0] ia2[2]-ia2[0]
                */
                
                const auto v0 = (ia1[1] - ia1[0]) * (ia2[2] - ia2[0]);
                const auto v1 = (ia2[1] - ia2[0]) * (ia1[2] - ia1[0]);
                return compute_sign_of_determinant(v1, v0);
        }

        // test ok
        /** compute the sign of the determinant of the form
                1 ia0
                1 ia1
         
         @param[in] ia0
         @param[in] ia1
         @retval      1 positive
         @retval     -1 negative
         @retval      0 zero(degenerate)
        */
        int compute_sign_of_determinant(T ia0, T ia1) const
        {
                if ( ia1 > ia0 ) {
                        return 1;
                } else if ( ia1 < ia0 ) {
                        return -1;
                } else {
                        return 0;
                }
        }

        // test ok
        bool judge(int sign, bool is_even_permutation) const
        {
                auto is_left = (sign == 1) ? true : false;
                return is_even_permutation ? is_left : !is_left;
        }

        
}; // class SymbolicPerturbation2d
#endif // SymbolicPerturbation2d_SymbolicPerturbation2d_h
